/* 
 * Copyright (C) 2013 Olivier Catoni and Thomas Mainguy
 * contact : catoni@noos.fr Thomas.Mainguy@ens.fr
 *
 * This file is part of ToricGrammars : a statistical learning model 
 * for linguistics
 *
 * ToricGrammars is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * ToricGrammars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


#ifndef TREES_H
#define TREES_H

#define MATHLIB_STANDALONE 1

#include <wx/wx.h>
#include <wx/stream.h>
#include <wx/wfstream.h>
#include <wx/txtstrm.h> 
#include <wx/string.h>
#include <wx/strconv.h>
#include <wx/filedlg.h>
#include <string>
#include <wx/tokenzr.h>
#include <wx/ffile.h>

#include <iostream>
#include <R_ext/Random.h>
#include <R.h>
#include <Rmath.h>


using namespace std;

const int LASTBIT=1;
const size_t BUFFLENGTH=1000000;
const int MAX_SENTENCE_LENGTH=100;

const size_t mask0 = 077;
const size_t mask1 = ~(7<<3);
const size_t mask2 = ~7;
const size_t mask7 = 7;

class Length {
	public:
	void setLength(size_t length) { m_size &= mask0; m_size |= (length << 6); }
	void setLeftOffset(size_t offset) { m_size &= mask1; m_size |= (offset << 3); }
	void setRightOffset(size_t offset) { m_size &= mask2; m_size |= offset; }
	unsigned int getLeftOffset() const { return (m_size>>3)&7; }
	unsigned int getRightOffset() const { return m_size&7; }
	size_t getLength() const { return m_size >> 6; } 
	size_t getStringLength() const { return (m_size >> 6) / sizeof(wxChar); } 
	size_t getBitLength() const { return (getLength()<<3) - getRightOffset(); } 
	private:
	size_t m_size;
};

class Node {
	public:
	Node(Length const &length, char * const &data);
	Node(Node const &node);
	Node * copySubTree() const;
	Node(wxString const & string);

	Node(wxChar const * const &array, size_t const &start, 
									size_t const &end, int const &count = 1);

	Node() : m_Son(), m_count(), m_length(), m_data() {}

	Node(wxFile &fileIn); 

  ~Node();  

	int draw(); 

	int getBit(size_t pos); 

	void cropLeft(size_t pos);

	void cropRight(size_t pos);

	void split(size_t pos); 

	size_t bitsCompare(Node const &node);

	void add(Node * node);

	void print(wxTextOutputStream & textOut); 
	void write(wxFile &fileOut); 
	Node * m_Son[2];
	int m_count[2];
	Length m_length;
	char *m_data;
};

/*
 *  The Tree<TNode> template class
 *  accommodates any type of nodes and organizes them into 
 *  a possibly incomplete binary tree.
 */
template <typename TNode> 
class Tree {
	public:

		Tree(): m_root(), m_pos(m_root) {
			// cout << "Tree()\n";
		}

		Tree(wxString const &filename): m_root(), m_pos(m_root) {
			// cout << "Tree(wxString const &)\n";
			wxFile fileIn(filename, wxFile::read);
			read(m_root, fileIn);
		}

		virtual ~Tree() {  
/*
 * We declare it virtual just in case we create derived classes from Tree
 */ 
			// cout << "~Tree()\n"; 
			delete m_root; 
		} 
		void print(wxTextOutputStream & textOut);
		/*
		 * Prints to a text stream
		 */
		void print(wxString const & filename) {
			wxFileOutputStream fileStream(filename);
			wxTextOutputStream textOut(fileStream);
			print(textOut);
		}

		/*
		 * Prints in text mode to a file a human readable representation
		 * of the tree.
		 */
		void insert(int index);
		/*
		 * Inserts at the given position. Creates new nodes if necessary.
		 * Just moves the m_pos pointer otherwise.
		 */
		int growTree();

		void write(wxString const &filename) {
			wxFile fileOut(filename, wxFile::write);
			write(fileOut);
		}
		  /*
			 * Writes a binary dump to a filename.
			 */
		void write(wxFile &fileOut) { 
				write(m_root, fileOut);
		}

  private:
		/*
		 * Now comes the low level recursive read. We test the pointers
		 * and restore them when they are not zero. Maybe we can start
		 * with recursive tree manipulation functions, after all. 
		 * They are so simple to write ! I tried by best to give 
		 * them a small footprint, passing arguments by const reference
		 * when possible.
		 */
		void read(TNode * &node, wxFile &fileIn); 
		  /* 
			 * And the internal recursive write
			 * function. Very simple : just writes the
			 * nodes in preorder. Passing const references
			 * is supposed to aid the compiler into optimizing
			 * function calls. 
			 */
		void write(TNode * const &node, wxFile &fileOut); 
		void print(wxTextOutputStream & textOut, TNode * const &node, int depth);
		friend class Label;
	protected:
		/*
		 * Let us protect the members.
		 */
		TNode *m_root;
		TNode *m_pos;
		int m_Count;
		int m_CrIndex;
};

const wxString whiteSpace = _(" \t\n\v\f\r");

// Class TreeNode derived from Tree<Node>, with added functionalities.

class TreeNode : public Tree<Node> {
	public:
	friend class Label;
	TreeNode() : Tree<Node>(), m_labelMax(), m_labelCount(1) {}
	TreeNode(TreeNode const &tree);
	TreeNode& operator=(TreeNode const & rightTree);
	TreeNode(wxString const &fileName) : Tree<Node>(fileName) {}
	virtual ~TreeNode() {};

	void setCount();
	static size_t last_of(wxChar const * const &array, size_t pos, size_t const &start, 
									wxString const & pattern);

	static size_t first_of(wxChar const * const &array, size_t pos, size_t const &end, 
									wxString const & pattern);

	static size_t last_not_of(wxChar const * const &array, size_t pos, 
									size_t const &start, 
									wxString const & pattern);

	static size_t first_not_of(wxChar const * const &array, size_t pos, 
									size_t const &end, 
									wxString const & pattern);

	static size_t prevTok(wxChar const * const &array, size_t const &pos, 
									size_t const &start, 
									wxString const & pattern = whiteSpace); 

	static size_t nextTok(wxChar const * const &array, size_t const &pos, 
									size_t const &end, 
									wxString const & pattern = whiteSpace); 

	static size_t insert(wxChar *array, size_t const &dStart, size_t const &dEnd, 
									size_t const &sEnd, size_t const &iStart, size_t const &iEnd);

	void add (wxString const & a_string); 

	bool merge(wxString const &oldLabel, wxString const &newLabel);

	void merge_all(wxString const &oldLabel, wxString const & newLabel);

	/*
	 * Selects the shortest context seen less than thresh times.
	 */
	wxString contLagSample(wxString const & starter, unsigned int thresh);
	wxString FBsample(wxString const & starter, int gen_cycles,
									int read_thresh, int fgen_thresh, int 
									bgen_thresh, bool global, wxString & hist);

	/*
	 *  Selects the largest context seen at least thresh times.
	 */

	wxString contextSample(wxString const & starter, unsigned int thresh); 

	wxString markovSample(wxString const &starter, int rank); 
	
	static bool cropString(wxString &string, int start, int end); 

	size_t sample_p(wxChar *array, size_t start, size_t end, 
									wxString const & pattern, wxString const & exclude = _("0"));

	size_t sample_p(wxChar *array, Node *a_node, size_t a_pos, 
									size_t a_end, wxString const &pattern, 
									wxString const & exclude);


	size_t sample(wxChar * array, size_t a_start, 
									size_t a_end, size_t a_newStart = BUFFLENGTH); 

	wxString sample(wxString const &string, size_t pos = 0); 

	int search_g(wxChar * array, size_t start, size_t end);

	void remove_all(wxChar *array, size_t start, size_t end);

	void remove_all(wxChar *array, size_t start, size_t end, int count);

	int remove_all(wxString const &string, int count = 1);

	void remove(wxChar *array, size_t start, size_t end, int a_count);

	int search(wxChar const * array, size_t start, size_t end); 

	int search(wxString const &string, size_t pos = 0); 

	void stringTree(); 

	void printS(wxString const & filename); 

	void printS(wxTextOutputStream &textOut, Node * const &node, 
									const int &pos); 	

	void readTextFB(wxString const &fileName, wxString const &starter = _("[0 "), 
									int weight=1); 

	void addPatterns(TreeNode &patterns, wxString const & starter = _("[0 "), 
									int weight=1);

	void add_all(wxChar *array, size_t start, size_t end, int const &count = 1);

	void add(wxChar *array, size_t start, size_t end, int const &count = 1);

	void add_all(wxString const &string, int count = 1);

	void readText(wxString const &fileName, int const &rank = 0); 

	wxString read(wxString starter);

	wxString read(wxChar const *a, size_t start, size_t end);

	void gatherNew(Node *node, size_t pos, TreeNode &oldExp,
									TreeNode &newExp);

	void gatherNew(wxString const &header, TreeNode &oldExp, 
									TreeNode &newExp);

	void read(Node *node, size_t pos, wxString &output);

	wxString sampleGrammar(wxString &hist);
	bool mergeGrammar();

	int m_labelCount, m_labelMax;
	protected:
	size_t m_offset;
	static char m_buff[];
};

class parseArg {
	public:
	parseArg() {}
	parseArg(Node *a_node): node(a_node), bitPos(), mid(), aIsOk(false),
										        aIsOut(false) { }
	Node *node;
	size_t bitPos;
	size_t mid;
	bool aIsOk;
	bool aIsOut;
	wxString leftLabel;
};

class splitArg {
	public:
	splitArg() {}
	splitArg(Node *a_node): node(a_node), bitPos(), mid(), aIsOk(false),
										        aIsOut(false) {}
	Node *node;
	size_t bitPos;
	size_t mid;
	bool aIsOk;
	bool aIsOut;
};


class Label {
	public:
	Label(TreeNode *tree, int maxOut, int maxIn);
	Label(TreeNode *tree, TreeNode *reference);
	Label(TreeNode *tree, wxString const & filename);
	void write(wxString const &filename);
	~Label();
	bool getLabel();
	void getLabel(Node *node, size_t a_pos, size_t a_end, int a_isOut);
	bool spreadLabel();
	bool spread(wxChar const *array, size_t start, size_t end);
	void maxIn(int const &max);
	void maxOut(int const &max);
	void parse(parseArg arg);
	void broadParse(parseArg arg);
	bool parse();
	void narrowParse(parseArg arg);
	bool narrowParse();
	void remap(Node *node, size_t pos, size_t charPos);
	bool remap();
	void addToRef(int maxRemap);
	void split(splitArg arg);
	bool split();
	void create(splitArg arg);
	bool create();
	TreeNode *m_tree;
	TreeNode *m_grammar;
	size_t m_count;
	int m_label;
	private:
	int m_max[2];
	
	wxChar *m_stem;
	size_t m_end;
	
	int m_isOut;
};

/*
 * A small insert function. Should be replaced by something 
 * more serious, where the index is not limited to 32 bits
 * and is stored in an array.
 */
template <>
void Tree<Node>::insert(int index); 

template <>
int Tree<Node>::growTree();

void saveSeed(wxString const &);
void getSeed(wxString const &);

#endif // TREES_H
